__version__='90.12.4.0.dev1'

import re
import pandas as pd
import numpy as np

def GraphDecomp(
    logfile # Logfile der Graphenzerlegung
    ):
    
    """
    Liest die SIR3S Graphenzerlegungs-logfile in ein Dictionary ein
    """

    f=open(logfile, 'r')
    lines = f.readlines()

    for line in lines:
        if re.search('^Nodes',line):
            idx_NODES = lines.index(line)
        elif re.search('^Links',line):
            idx_LINKS = lines.index(line)


    df_NODES = pd.DataFrame(np.row_stack([row.split(';') for row in lines[idx_NODES+2:idx_LINKS-1]]))
    df_NODES = df_NODES.iloc[: , :-1]

    df_NODES_ColNames = lines[idx_NODES+1].split(';')
    for x in range(len(df_NODES_ColNames)):
        df_NODES_ColNames[x] = df_NODES_ColNames[x].strip()

    df_NODES.columns = df_NODES_ColNames


    df_LINKS = pd.DataFrame(np.row_stack([row.split(';') for row in lines[idx_LINKS+2:]]))

    df_LINKS_ColNames = lines[idx_LINKS+1].split(';')
    for x in range(len(df_LINKS_ColNames)):
        df_LINKS_ColNames[x] = df_LINKS_ColNames[x].strip()

    df_LINKS[df_LINKS.columns[-1]] = df_LINKS[df_LINKS.columns[-1]].apply(lambda x: x.strip())

    df_LINKS.columns = df_LINKS_ColNames


    GraphDecomp = {'NODES':df_NODES, 'LINKS':df_LINKS}
    return GraphDecomp


__version__='90.12.4.0.dev1'

import os
import fitz
from PIL import Image

NDD_left = 220
NDD_top = 60
NDD_right = 4650
NDD_bottom = 3230

NFD_left = 220
NFD_top = 220
NFD_right = 3090
NFD_bottom = 3255

Station_left = 220
Station_top = 220
Station_right = 4425
Station_bottom = 3255

ZK_left = 70
ZK_top = 230
ZK_right = 3230
ZK_bottom = 2260

def fPngFromPdf(
    file # PDF-File
   ,crop='ZK' # how to crop PNG-File after extraction: 'ZK', 'NDD', 'Station', 'NFD' or None or (left,top,right,bottom)
    ):
    """
    generates PNG-Files for every Page in PDF-File

    cropping options:

        * ZK
        * NDD
        * Station
        * NFD
        * None
        * (left, top, right, bottom)
    """

    dirname=os.path.dirname(file)
    dummy,filename=os.path.split(file)
    filenameNetto,fileExt=os.path.splitext(filename)


    zoom_x = 4.0  # horizontal zoom
    zoom_y = 4.0  # vertical zoom
    mat = fitz.Matrix(zoom_x, zoom_y)  # zoom factor 2 in each dimension


    with fitz.open(file)as doc:  # open document

        if crop=='ZK':
            (crop_left,crop_top,crop_right,crop_bottom)=(ZK_left,ZK_top,ZK_right,ZK_bottom)
        elif crop=='NDD':
            (crop_left,crop_top,crop_right,crop_bottom)=(NDD_left,NDD_top,NDD_right,NDD_bottom)
        elif crop=='NFD':
            (crop_left,crop_top,crop_right,crop_bottom)=(NFD_left,NFD_top,NFD_right,NFD_bottom)
        elif crop=='Station':
            (crop_left,crop_top,crop_right,crop_bottom)=(Station_left,Station_top,Station_right,Station_bottom)
        elif crop == None:
            pass
        else:
            (crop_left,crop_top,crop_right,crop_bottom)=crop

        for idxPage,page in enumerate(doc):  # iterate through the pages

            # Ausgabedatei
            if len(doc)>1:
                png_name="{:s}_{:d}.png".format(filenameNetto,idxPage+1)
            else:
                png_name="{:s}.png".format(filenameNetto)
            png_NameFull=os.path.join(dirname,png_name)

            pix = page.get_pixmap(matrix=mat)  # render page to an image
            pix.save(png_NameFull)  # store image as a PNG

            # Nachbearbeitung
            if crop != None:
                im = Image.open(png_NameFull)
                im = im.crop((crop_left,crop_top,crop_right,crop_bottom))
                im.save(png_NameFull)



import os
import sys

import re

import pyodbc

import logging

logger = logging.getLogger('PT3S') 


class Error(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class accdb(object):
    """
    create some SIR 3S related Views in an Access-DB
    """

    def __init__(self,accFile):

        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 
        
        try:             
            if os.path.exists(accFile):  
                if os.access(accFile,os.R_OK):
                    if os.access(accFile,os.W_OK):
                        pass # OK
                    else:
                        logStrFinal="{:s}accFile: {:s}: Not writetable.".format(logStr,accFile) 
                        raise Error(logStrFinal)
            else:                
                logStrFinal="{:s}accFile: {:s}: Not readable!".format(logStr,accFile)     
                raise Error(logStrFinal)  
                                      
            # die MDB existiert und ist lesbar
            logger.debug("{:s}accFile (abspath): {:s} is read- and writetable.".format(logStr,os.path.abspath(accFile))) 
           
            Driver=[x for x in pyodbc.drivers() if x.startswith('Microsoft Access Driver')]
            if Driver == []:
                logStrFinal="{:s}{:s}: No Microsoft Access Driver!".format(logStr,accFile)     
                raise Error(logStrFinal)  

            # ein Treiber ist installiert
            self.conStr=(
                r'DRIVER={'+Driver[0]+'};'
                r'DBQ='+os.path.abspath(accFile)+';'
                )
            logger.debug("{:s}conStr: {:s}".format(logStr,self.conStr)) 

            # Verbindung ...
            #self.con = pyodbc.connect(self.conStr)
            #self.cur = self.con.cursor()

            ## all Tables in DB
            #tableNames=[table_info.table_name for table_info in cur.tables(tableType='TABLE')]
            #logger.debug("{0:s}tableNames: {1:s}".format(logStr,str(tableNames))) 
            #allTables=set(tableNames)
                     
        except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise Error(logStrFinal)              
        finally:
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     

    def executeSQLFile(self,sqlFile):

        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 
        
        try:    

            if os.path.exists(sqlFile):  
                if os.access(sqlFile,os.R_OK):
                    pass
                else:
                    logStrFinal="{:s}sqlFile: {:s}: Not readable.".format(logStr,sqlFile)
                    raise Error(logStrFinal)  
            else:
                logStrFinal="{:s}sqlFile: {:s}: Not existing.".format(logStr,sqlFile)
                raise Error(logStrFinal)  

            logger.debug("{:s}sqlFile (abspath): {:s} is readable.".format(logStr,os.path.abspath(sqlFile))) 
           
            con = pyodbc.connect(self.conStr)
            cur = con.cursor()            
            tableNames=[table_info.table_name for table_info in cur.tables(tableType='TABLE')]
            viewNames=[table_info.table_name for table_info in cur.tables(tableType='VIEW')]
            allNames=tableNames+viewNames
            allNames=set(allNames)
            allNames=list(allNames)
            cur.close()
            con.close()

            sqlfile=open(sqlFile, 'r')
            lines=sqlfile.read()
            sqlfile.close()

            for stmt in self._getSQLStmts(lines):

                logger.debug("{:s}stmt: {:s}".format(logStr,stmt))    
                                           
                con = pyodbc.connect(self.conStr)
                cur = con.cursor()

                try:
                    cur.execute(stmt)#+';')   
                    
                except Exception as e:
                    #logger.debug("{:s}ERROR: Stmt: {:s}".format(logStr,stmt))  
                    logger.debug("{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e)))
                    
                finally:
                    con.commit()
                    cur.close()
                    con.close()
                        
                    
        except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise Error(logStrFinal)              
        finally:
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     


    def _getSQLStmts(self,text):
        current = ''
        state = None
        for c in text:
            if state is None:  # default state, outside of special entity
                current += c
                if c in '"\'':
                    # quoted string
                    state = c
                elif c == '-':
                    # probably "--" comment
                    state = '-'
                elif c == '/':
                    # probably '/*' comment
                    state = '/'
                elif c == ';':
                    # remove it from the statement
                    current = current[:-1].strip()
                    ##current = current.strip()
                    # and save current stmt unless empty
                    if current:
                        yield current
                    current = ''
            elif state == '-':
                if c != '-':
                    # not a comment
                    state = None
                    current += c
                    continue
                # remove first minus
                current = current[:-1]
                # comment until end of line
                state = '--'
            elif state == '--':
                if c == '\n':
                    # end of comment
                    # and we do include this newline
                    current += c
                    state = None
                # else just ignore
            elif state == '/':
                if c != '*':
                    state = None
                    current += c
                    continue
                # remove starting slash
                current = current[:-1]
                # multiline comment
                state = '/*'
            elif state == '/*':
                if c == '*':
                    # probably end of comment
                    state = '/**'
            elif state == '/**':
                if c == '/':
                    state = None
                else:
                    # not an end
                    state = '/*'
            elif state[0] in '"\'':
                current += c
                if state.endswith('\\'):
                    # prev was backslash, don't check for ender
                    # just revert to regular state
                    state = state[0]
                    continue
                elif c == '\\':
                    # don't check next char
                    state += '\\'
                    continue
                elif c == state[0]:
                    # end of quoted string
                    state = None
            else:
                raise Exception('Illegal state %s' % state)

        if current:
            current = current.rstrip(';').strip()
            if current:
                yield current


"""

"""

__version__='90.12.4.0.dev1'

import warnings
warnings.filterwarnings("ignore")

import sys
import logging

import pandas as pd
import geopandas

import matplotlib.pyplot as plt
import matplotlib
from matplotlib import cm

import contextily as cx

logger = logging.getLogger('PT3S')  
if __name__ == "__main__":
    logger.debug("{0:s}{1:s}".format('in MODULEFILE: __main__ Context:','.')) 
else:
    logger.debug("{0:s}{1:s}{2:s}{3:s}".format('in MODULEFILE: Not __main__ Context: ','__name__: ',__name__," ."))

class NFDError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class NFD():

    """
    Raises:
        NFDError
    """

    def __init__(self):
        self.name = 'NFD'

    def plotNFD(gdf_FWVB
                ,gdf_ROHR
                ,colors_FWVB = ['green', 'red']
                ,colors_ROHR = ['lightgray', 'dimgray']
                ,norm_min_FWVB = 24
                ,norm_max_FWVB = 38
                ,yTicks = None
                ,yTickLabels = None
                ,norm_min_ROHR = 333
                ,norm_max_ROHR = 666
                ,FWVB_edgecolor = 'White'
                ,FWVB_groesse = 0.1
                ,color_FWVB_under = 'green'
                ,color_FWVB_over = 'red'
                ,color_ROHR_under = 'magenta'
                ,color_ROHR_over = 'lime'
                ,Anzeigen_ROHR = True
                ,ROHR_Ampel = False
                ,ROHR_dicke = 0.0025
                ,column_ROHR_Dicke = 'DI'
                ,color_ROHR = 'grey'
                ,LabelsOnTop = True
                ,FWVB_markerscaling_col = 'W0'
                ,FWVB_cmap_col = 'ZKOR_k'
                ,Map_resolution = 12
                ,FWVB_cbar_label = 'default'
                ,cbar_size = 0.7
                ):

        """
        Die Funktion plottet FWVB und wahlweise Rohre mit Hintergrundkarte.
        Groesse und Farbe der FWVB/Rohre koennen an eigen gewaehlte Spalten angepasst werden.

        Args:

            * gdf_FWVB:               DataFrame der FWVB (muss vor dem Ausfuehren dieser Funktion definiert werden)
            * gdf_ROHR:               DataFrame der ROHRE (muss vor dem Ausfuehren dieser Funktion definiert werden)
            * colors_FWVB:            Bsp.: ['green', 'red'] Colormap der FWVB (niedrige Werte beginnen bei linker Farbe(hier 'green'))
            * colors_ROHR:            Bsp.: ['lightgray', 'dimgray'] Colormap der Rohre (niedrige Werte beginnen bei linker Farbe(hier 'dimgray'))
            * norm_min_FWVB:          Unterer Wert der Farbskala fuer FWVB
            * norm_max_FWVB:          Oberer Wert der Farbskala fuer FWVB
            * yTicks:                 Hier koennen die yTicks der Colorbar angegeben werden
            * yTickLabels:            Hier koennen die yTickLabels der Colorbar angegeben werden
            * norm_min_ROHR:          Unterer Wert der Farbskala fuer Rohre
            * norm_max_ROHR:          Oberer Wert der Farbskala fuer Rohre
            * FWVB_edgecolor:         Randfarbe der FWVB
            * FWVB_groesse:           Faktor, der die Groesse der FWVB zusaetzlich skaliert
            * color_FWVB_under:       Farbe der FWVB mit Werten unterhalb der definierten Farbskala
            * color_FWVB_over:        Farbe der FWVB mit Werten ueberhalb der definierten Farbskala
            * color_ROHR_under:       Farbe der Rohre mit Werten unterhalb der definierten Farbskala
            * color_ROHR_over:        Farbe der Rohre mit Werten ueberhalb der definierten Farbskala
            * Anzeigen_ROHR:          Bestimmt, ob Rohre angezeigt werden (True/False)
            * ROHR_Ampel:             Sollen die Rohre eine Farbampel besitzen (True/False)
            * ROHR_dicke:             Faktor, der die Dicke der Rohre zusaetzlich skaliert
            * column_ROHR_Dicke:      Spalte, welche die Dicke der Rohre definiert
            * color_ROHR:             Farbe der Rohre (falls nicht in Ampeldarstellung)
            * LabelsOnTop:            Sollen die Labels ueber dem geplotteten angezeigt werden (True/False)
            * FWVB_markerscaling_col: Gibt die Spalte der FWVB an, mit der die Marker (markersize, linewidth, alpha) skaliert werden sollen
            * FWVB_cmap_col:          Gibt die Spalte der FWVB an, mit der diese in die Colormap eingeordnet werden
            * Map_resolution:         0-20 Aufloesung der Hintergrundkarte (Je hoeher das Zommlevel, desto hoeher die Aufloesung)
            * FWVB_cbar_label:        Titel der Colorbar fuer die direkten FWVB
            * cbar_size:              Faktor mit dem die groesse der Colorbar skaliert wird
        """

        logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.'))


        fig = plt.gcf()
        ax = plt.gca()

        norm_FWVB_color = plt.Normalize(vmin=norm_min_FWVB, vmax=norm_max_FWVB)

        norm_ROHR = plt.Normalize(vmin=norm_min_ROHR, vmax=norm_max_ROHR)

        cmap_FWVB = matplotlib.colors.LinearSegmentedColormap.from_list('FWVB', colors_FWVB, N = 256) #Erstellen der Colormap fuer die FWVB
        cmap_FWVB.set_under(color_FWVB_under)
        cmap_FWVB.set_over(color_FWVB_over)

        cmap_ROHR = matplotlib.colors.LinearSegmentedColormap.from_list('ROHR', colors_ROHR, N = 256) #Erstellen der Colormap fuer die Rohre
        cmap_ROHR.set_under(color_ROHR_under)
        cmap_ROHR.set_over(color_ROHR_over)

        norm_FWVB_size = plt.Normalize(vmin = gdf_FWVB[FWVB_markerscaling_col].min(), vmax = gdf_FWVB[FWVB_markerscaling_col].max()) #Festlegen wie diese Spalte normiert werden soll

        try:
            gdf_FWVB.plot(ax = ax
                        ,marker = '.'
                        ,markersize = gdf_FWVB[FWVB_markerscaling_col] * FWVB_groesse #Markersize ist abhaengig von der zum Markerscaling ausgewaehlten Spalte und einem Faktor
                        ,c = cmap_FWVB(norm_FWVB_color(gdf_FWVB[FWVB_cmap_col])) #Einfaerben der Marker entsprechend der definierten Colormap und ausgewaehlten Spalte
                        ,edgecolor = FWVB_edgecolor #FWVB Randfarbe des Markers
                        ,linewidth = norm_FWVB_size(gdf_FWVB[FWVB_markerscaling_col]) / 2 #FWVB Randdicke, abhaengig von der zum Markerscaling ausgewaehlten Spalte. Die Randdicke wird im Anschluss nochmal durch 2 geteilt.
                        ,alpha =  0.9 - (norm_FWVB_size(gdf_FWVB[FWVB_markerscaling_col]) / 2) #FWVB Durchsichtigkeit, abhaengig von der zum Markerscaling ausgewaehlten Spalte. Mit dem Teilen durch 2 ist die maximale Durchsichtigkeit bei 0.5 (0 bei Kleinen-0.5 bei Grossen). Um dies umzudrehen zieht man von 1 ab. Hier wird von 0.9 abgezogen. Damit sind kleinsten FWVB immer noch leicht durchsichtig und die Groessten trotzdem noch erkennbar (0.9 bei Kleinen - 0.4 bei Grossen) 
                        ,zorder = 2) #Reihenfolge zu den anderen geploteten Objekten (je hoeher die Zahl, desto hoeher die geplottete Ebene; Hoehe im Sinne von Vorder- und Hintergrund)

            if Anzeigen_ROHR:

                if ROHR_Ampel:
                    gdf_ROHR.plot(ax = ax
                                    ,zorder = 1
                                    ,linewidth = gdf_ROHR[column_ROHR_Dicke] * ROHR_dicke
                                    ,c = cmap_ROHR(norm_ROHR(gdf_ROHR[column_ROHR_Dicke])))

                else:    
                    gdf_ROHR.plot(ax = ax
                                    ,zorder = 1
                                    ,linewidth = gdf_ROHR[column_ROHR_Dicke] * ROHR_dicke
                                    ,color = color_ROHR)

            if LabelsOnTop:
                cx.add_basemap(ax, crs=gdf_FWVB.crs.to_string(), source = cx.providers.CartoDB.PositronNoLabels, zoom = Map_resolution)
                cx.add_basemap(ax, crs=gdf_FWVB.crs.to_string(), source = cx.providers.CartoDB.PositronOnlyLabels, zoom = Map_resolution)
            else:
                cx.add_basemap(ax, crs=gdf_FWVB.crs.to_string(), source = cx.providers.CartoDB.Positron, zoom = Map_resolution)


            cbar=fig.colorbar(cm.ScalarMappable(norm = norm_FWVB_color, cmap = cmap_FWVB)
                        ,ax = ax
                        ,label = FWVB_cbar_label
                        ,shrink = cbar_size
                        ,ticks=yTicks
                        ,pad = 0.01)
            if yTickLabels != None:
                cbar.ax.set_yticklabels(yTickLabels)

            plt.axis('off')

        except NFDError:
            raise            
        except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise NFDError(logStrFinal)                       
        finally:       
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))
            return


"""
Testtreiber.

"""

import warnings # 3.6
#...\Anaconda3\lib\site-packages\h5py\__init__.py:36: FutureWarning: Conversion of the second argument of issubdtype from `float` to `np.floating` is deprecated. In future, it will be treated as `np.float64 == np.dtype(float).type`.
#   from ._conv import register_converters as _register_converters
#...\PT3S\Mx.py:1: FutureWarning: pandas.tslib is deprecated and will be removed in a future version.
#   You can access Timestamp as pandas.Timestamp
warnings.simplefilter(action='ignore', category=FutureWarning)

import os
import sys

import logging
logger = logging.getLogger('PT3S')  

if __name__ == "__main__":
    logger.debug("{0:s}{1:s}".format('in MODULEFILE: __main__ Context: ',' .')) 
else:
    logger.debug("{0:s}{1:s}{2:s}{3:s}".format('in MODULEFILE: Not __main__ Context: ','__name__: ',__name__," .")) 

# ---
# --- main Imports
# ---
import argparse
import unittest
import doctest
import re

if __name__ == "__main__":
    """
    Run Tests.
    """

    try:      
        

                                      
        # Arguments      
        parser = argparse.ArgumentParser(description='Run Tests.'
        ,epilog='''
        UsageExamples: 

        Modultest:

        -q -m 1 -t both -w OneLPipe -w LocalHeatingNetwork -w GPipe -w GPipes -w TinyWDN 

        Singletests:
        
        -q -m 0 -s "^Mx\."  -s "^Xm\." -t both -y yes -z no -w OneLPipe -w LocalHeatingNetwork -w GPipe -w GPipes -w TinyWDN 

        Singletests: separater MockUp-Lauf:

        -q -m 0 -t before -u yes -w DHNetwork
        
        Singletests (die auf dem vorstehenden MockUp-Lauf basieren):
        -q -m 0 -s "^Rm\." -z no -w DHNetwork

        '''                                 
        )

        group = parser.add_mutually_exclusive_group()                                
        group.add_argument("-v","--verbose", help="Debug Messages On", action="store_true",default=True)      
        group.add_argument("-q","--quiet", help="Debug Messages Off", action="store_true")           

        parser.add_argument('--testDir',type=str,default='testdata',help="value for global 'testDir' i.e. testdata")
        parser.add_argument('--dotResolution',type=str,default='.1',help="value for global 'dotResolution' i.e. .1 (default); use NONE for no dotResolution")      
                                 
        parser.add_argument("-m","--moduleTest", help="execute the Module Doctest On/Off: -m 1 (default)", action="store",default='1')      
        parser.add_argument("-s","--singleTest", help='execute single Doctest: Exp.1: -s  "^Rm.": all Doctests in Module Rm are executed - but not the Module Doctest (which is named Rm) Exp.2:  -s "^Xm."  -s "^Mx."  -s "^Rm.": all Doctests in the 3 Modules are executed - but not the Module Doctests'
                            ,action="append"
                            ,default=[])       
        
        parser.add_argument("-x","--singleTestNO", help='execute NOT single Doctest: Exp.1: -s  "^Rm.": NO Doctests in Module Rm are executed - but not the Module Doctest (which is named Rm) Exp.2:  -s "^Xm."  -s "^Mx."  -s "^Rm.": NO Doctests in the 3 Modules are executed - but not the Module Doctests'
                            ,action="append"
                            ,default=[])       
        
        parser.add_argument("-t","--delGenFiles", help="Tests: decide if generated Files - i.e. .h5-Files - shall be deleted: Exp.: -t both: generated Files are deleted before and after the Tests"
                            ,choices=['before', 'after', 'both','nothing'],default='nothing')

        parser.add_argument("-y","--mockUpDetail1", help="MockUp Detail1: decide if NoH5 shall be used during MockUps: Exp.: -y yes"
                            ,choices=['no','yes'],default='no')

        parser.add_argument("-z","--mockUpDetail2", help="MockUp Detail2: decide if Sync/Add and ToH5 shall be done during MockUps: Exp.: -z no"
                            ,choices=['no','yes'],default='yes')

        parser.add_argument("-u","--mockUpAtTheEnd", help="Tests: decide if (after all Tests and after delGenFiles) some mockUp shall be done: Exp.: -u yes"
                            ,choices=['no','yes'],default='no')

        parser.add_argument("-w","--testModel", help='specify a testModel: Exp.: -w DHNetwork'
                            ,action="append"
                            ,default=[])           

        parser.add_argument("-l","--logExternDefined", help="Logging (File etc.) ist extern defined", action="store_true",default=False)      

        logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)

        args = parser.parse_args()


        class LogStart(Exception):
            pass

        

        try:
            #logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
            if args.logExternDefined:  
                logger = logging.getLogger('PT3S')  
                logStr=logStr+" (Logging extern defined) "
            else:                
                logFileName = 'PT3S.log' 
        
                loglevel = logging.INFO
                logging.basicConfig(filename=logFileName
                                    ,filemode='w'
                                    ,level=loglevel
                                    ,format="%(asctime)s ; %(name)-60s ; %(levelname)-7s ; %(message)s")    

                fileHandler = logging.FileHandler(logFileName)        
                logger.addHandler(fileHandler)

                consoleHandler = logging.StreamHandler()
                consoleHandler.setFormatter(logging.Formatter("%(levelname)-7s ; %(message)s"))
                consoleHandler.setLevel(logging.INFO)
                logger.addHandler(consoleHandler)
                          
            raise LogStart
        except LogStart:   
            if args.verbose:  # default         
                logger.setLevel(logging.DEBUG)  
            if args.quiet:    # Debug Messages are turned Off
                logger.setLevel(logging.ERROR)  
                args.verbose=False            
            logger.debug("{0:s}{1:s}".format(logStr,'Start.'))             
        else:
            pass
                                            
        logger.debug("{0:s}{1:s}{2:s}".format(logStr,'Argumente:',str(sys.argv))) 
        logger.debug("{0:s}{1:s}{2:s}".format(logStr,'testDir: ',args.testDir)) 

        if args.dotResolution == 'NONE':
            args.dotResolution=''

        try:
            from PT3S import Mx, Xm, Rm
        except ImportError:
            logger.debug("{0:s}{1:s}".format("test: from PT3S import Mx, Xm, Rm: ImportError: ","trying import Mx, Xm, Rm ..."))  
            import Mx, Xm, Rm

        testModels=args.testModel 
        if len(testModels)==0:
            testModels=['OneLPipe','LocalHeatingNetwork','GPipe','GPipes','TinyWDN','DHNetwork']

        # die Modultests gehen i.d.R. vom Ausgangszustand aus; Relikte aus alten Tests müssen daher i.d.R. gelöscht werden ...
        # die Modultests erfahren kein MockUp
        if args.delGenFiles in ['before','both']:
                for testModel in testModels:   
                    #Xm                    
                    xmlFile=os.path.join(os.path.join('.',args.testDir),testModel+'.XML')                      
                    h5FileXm=os.path.join(os.path.join('.',args.testDir),testModel+'.h5')
                    #Mx
                    mx1File=os.path.join('.',os.path.join(args.testDir,'WD'+testModel+'\B1\V0\BZ1\M-1-0-1'+args.dotResolution+'.MX1'))                     
                    (wD,fileName)=os.path.split(mx1File)
                    (base,ext)=os.path.splitext(fileName)
                    (base,dotResolution)=os.path.splitext(base)                                                                             
                    h5File=wD+os.path.sep+base+'.'+'h5'    
                    h5FileVecs=wD+os.path.sep+base+dotResolution+'.'+'vec'+'.'+'h5' 
                    h5FileMx1FmtString=h5File+'.metadata'
                    #loeschen
                    for file in [h5FileXm,h5File,h5FileVecs,h5FileMx1FmtString]:                    
                        if os.path.exists(file):      
                            logger.debug("{:s}Tests Vorbereitung {:s} Delete {:s} ...".format(logStr,testModel,file)) 
                            os.remove(file)


        if args.moduleTest == '1':
            # as unittests
            logger.info("{0:s}{1:s}{2:s}".format(logStr,'Start unittests (by DocTestSuite...). testDir: ',args.testDir)) 

            dtFinder=doctest.DocTestFinder(recurse=False,verbose=args.verbose) # recurse = False findet nur den Modultest


            suite=doctest.DocTestSuite(Mx,test_finder=dtFinder #,setUp=setUpFct
                                   ,globs={'testDir':args.testDir
                                           ,'dotResolution':args.dotResolution
                                           })   
            unittest.TextTestRunner().run(suite)

            suite=doctest.DocTestSuite(Xm,test_finder=dtFinder #,setUp=setUpFct
                                   ,globs={'testDir':args.testDir
                                           ,'dotResolution':args.dotResolution
                                           })   
            unittest.TextTestRunner().run(suite)

            suite=doctest.DocTestSuite(Rm,test_finder=dtFinder #,setUp=setUpFct
                                   ,globs={'testDir':args.testDir
                                           ,'dotResolution':args.dotResolution
                                           })   
            unittest.TextTestRunner().run(suite)
                      
            # as doctests
            logger.info("{0:s}{1:s}{2:s}".format(logStr,'Start doctests. testDir: ',args.testDir)) 

            dtRunner=doctest.DocTestRunner(verbose=False) 

            dTests=dtFinder.find(Mx,globs={'testDir':args.testDir,'dotResolution':args.dotResolution}) 
            dtRunner.run(dTests[0])

            dTests=dtFinder.find(Xm,globs={'testDir':args.testDir,'dotResolution':args.dotResolution})
            dtRunner.run(dTests[0])

            dTests=dtFinder.find(Rm,globs={'testDir':args.testDir,'dotResolution':args.dotResolution}) 
            dtRunner.run(dTests[0])

        if len(args.singleTest)>0:

            #Relikte, die die Modultests oder andere Tests produziert haben ggf. loeschen
            #diese Relikte können z.B. aus den Modultests stammen wenn diese vorher durchgeführt wurden und diese "nicht aufräumen"
            if args.delGenFiles in ['before','both']:
                for testModel in testModels:   
                    #Xm                    
                    xmlFile=os.path.join(os.path.join('.',args.testDir),testModel+'.XML')                      
                    h5FileXm=os.path.join(os.path.join('.',args.testDir),testModel+'.h5')
                    #Mx
                    mx1File=os.path.join('.',os.path.join(args.testDir,'WD'+testModel+'\B1\V0\BZ1\M-1-0-1'+args.dotResolution+'.MX1'))                     
                    (wD,fileName)=os.path.split(mx1File)
                    (base,ext)=os.path.splitext(fileName)
                    (base,dotResolution)=os.path.splitext(base)                                                                             
                    h5File=wD+os.path.sep+base+'.'+'h5'    
                    h5FileVecs=wD+os.path.sep+base+dotResolution+'.'+'vec'+'.'+'h5' 
                    h5FileMx1FmtString=h5File+'.metadata'
                    #loeschen
                    for file in [h5FileXm,h5File,h5FileVecs,h5FileMx1FmtString]:                    
                        if os.path.exists(file):      
                            logger.debug("{:s}singleTests Vorbereitung {:s} Delete {:s} ...".format(logStr,testModel,file)) 
                            os.remove(file)

            #MockUp
          
            logger.debug("{:s}singleTests Vorbereitung Start ...".format(logStr)) 
            xms={}   
            mxs={} 
            ms={}
                               
            for testModel in testModels:   
                logger.debug("{:s}singleTests Vorbereitung {:s} Start ...".format(logStr,testModel)) 

                #Xm
                xmlFile=os.path.join(os.path.join('.',args.testDir),testModel+'.XML')  
                ms[testModel]=xmlFile
                if args.mockUpDetail1 in ['yes']:   
                    xm=Xm.Xm(xmlFile=xmlFile,NoH5Read=True) 
                else:
                    xm=Xm.Xm(xmlFile=xmlFile)      
                logger.debug("{:s}singleTests Vorbereitung {:s} xm instanziert.".format(logStr,testModel)) 

                #Mx
                mx1File=os.path.join('.',os.path.join(args.testDir,'WD'+testModel+'\B1\V0\BZ1\M-1-0-1'+args.dotResolution+'.MX1'))    
                if args.mockUpDetail1 in ['yes']:   
                     mx=Mx.Mx(mx1File=mx1File,NoH5Read=True) 
                else:
                    mx=Mx.Mx(mx1File=mx1File) 
                logger.debug("{:s}singleTests Vorbereitung {:s} mx instanziert.".format(logStr,testModel)) 

                if args.mockUpDetail2 in ['yes']:                    
                    #Sync                
                    xm.MxSync(mx=mx)
                    xm.MxAdd(mx=mx)
                    logger.debug("{:s}singleTests Vorbereitung {:s} Sync/Add erfolgt.".format(logStr,testModel)) 

                    #H5
                    xm.ToH5()
                    mx.ToH5()
                    logger.debug("{:s}singleTests Vorbereitung {:s} ToH5 erfolgt.".format(logStr,testModel)) 
                
                xms[testModel]=xm
                mxs[testModel]=mx
                logger.debug("{:s}singleTests Vorbereitung {:s} fertig.".format(logStr,testModel)) 
 
            # Tests suchen
            dtFinder=doctest.DocTestFinder(verbose=args.verbose)
            
            logger.debug("{:s}Tests suchen in Mx ...".format(logStr)) 
            dTests=dtFinder.find(Mx,globs={'testDir':args.testDir
                                           ,'dotResolution':args.dotResolution
                                           ,'mxs':mxs}) 

            logger.debug("{:s}Tests suchen in Xm ...".format(logStr)) 
            dTests.extend(dtFinder.find(Xm,globs={'testDir':args.testDir                                                                                    
                                           ,'xms':xms
                                           ,'ms':ms})) 

            logger.debug("{:s}Tests suchen in Rm ...".format(logStr)) 
            dTests.extend(dtFinder.find(Rm,globs={'testDir':args.testDir     
                                           ,'dotResolution':args.dotResolution
                                           ,'xms':xms
                                           ,'mxs':mxs})) 

            for test in dTests:
                pass
                #logger.debug("{0:s}singleTests: {1:s}: {2:s} ...".format(logStr,'Test found',test.name)) 

            # gefundene Tests mit geforderten Tests abgleichen

            testsToBeExecuted=[]
            for expr in args.singleTest:
                logger.debug("{0:s}singleTests: {1:s}: {2:s} ...".format(logStr,'Searching in Tests found for Expr     TBD',expr.strip("'")))                
                testsToBeExecuted=testsToBeExecuted+[test for test in dTests if re.search(expr.strip("'"),test.name) != None]     
            logger.debug("{0:s}singleTests: {1:s}: {2:s}".format(logStr,'    TBD',str(sorted([test.name for test in testsToBeExecuted]))))                   

            testsNotToBeExecuted=[]
            for expr in args.singleTestNO:
                logger.debug("{0:s}singleTests: {1:s}: {2:s} ...".format(logStr,'Searching in Tests found for Expr NOT TBD',expr.strip("'")))      
                testsNotToBeExecuted=testsNotToBeExecuted+[test for test in testsToBeExecuted if re.search(expr.strip("'"),test.name) != None]       
            logger.debug("{0:s}singleTests: {1:s}: {2:s}".format(logStr,'NOT TBD',str(sorted([test.name for test in testsNotToBeExecuted]))))    

            # effektiv auszuführende Tests 
            testsToBeExecutedEff=sorted(set(testsToBeExecuted)-set(testsNotToBeExecuted),key=lambda test: test.name)
            
            dtRunner=doctest.DocTestRunner(verbose=args.verbose) 
            for test in testsToBeExecutedEff:                      
                    logger.debug("{0:s}singleTests: {1:s}: {2:s} ...".format(logStr,'Running Test',test.name)) 
                    dtRunner.run(test)                     

        if args.delGenFiles in ['after','both']:          
            for testModel in testModels:              
                    #Xm                    
                    xmlFile=os.path.join(os.path.join('.',args.testDir),testModel+'.XML')                      
                    h5FileXm=os.path.join(os.path.join('.',args.testDir),testModel+'.h5')
                    #Mx
                    mx1File=os.path.join('.',os.path.join(args.testDir,'WD'+testModel+'\B1\V0\BZ1\M-1-0-1'+args.dotResolution+'.MX1'))                     
                    (wD,fileName)=os.path.split(mx1File)
                    (base,ext)=os.path.splitext(fileName)
                    (base,dotResolution)=os.path.splitext(base)                                                                             
                    h5File=wD+os.path.sep+base+'.'+'h5'    
                    h5FileVecs=wD+os.path.sep+base+dotResolution+'.'+'vec'+'.'+'h5' 
                    h5FileMx1FmtString=h5File+'.metadata'
                    #loeschen
                    for file in [h5FileXm,h5File,h5FileVecs,h5FileMx1FmtString]:                    
                        if os.path.exists(file):      
                            logger.debug("{:s}Tests Nachbereitung {:s} Delete {:s} ...".format(logStr,testModel,file)) 
                            os.remove(file)

        if args.mockUpAtTheEnd in ['yes']:                
            for testModel in testModels:
                logger.debug("{:s}Tests Nachbereitung {:s} mockUpAtTheEnd ...".format(logStr,testModel)) 

                #Mx
                mx1File=os.path.join('.',os.path.join(args.testDir,'WD'+testModel+'\B1\V0\BZ1\M-1-0-1'+args.dotResolution+'.MX1')) 
                if args.mockUpDetail1 in ['yes']:                 
                    mx=Mx.Mx(mx1File=mx1File,NoH5Read=True) # avoid doing anything than just plain Init                          
                else:
                    mx=Mx.Mx(mx1File=mx1File) 
                #Xm                     
                xmlFile=os.path.join(os.path.join('.',args.testDir),testModel+'.XML')
                if args.mockUpDetail1 in ['yes']:    
                    xm=Xm.Xm(xmlFile=xmlFile,NoH5Read=True) # avoid doing anything than just plain Init                                           
                else:
                    xm=Xm.Xm(xmlFile=xmlFile) 
                   
                if args.mockUpDetail2 in ['yes']:                    
                    #Sync                
                    xm.MxSync(mx=mx)
                    xm.MxAdd(mx=mx)
                    logger.debug("{:s}Tests Nachbereitung {:s} Sync/Add erfolgt.".format(logStr,testModel)) 

                    #H5
                    xm.ToH5()
                    mx.ToH5()
                    logger.debug("{:s}Tests Nachbereitung {:s} ToH5 erfolgt.".format(logStr,testModel)) 
                                
    except SystemExit:
        pass                                              
    except:
        logger.error("{0:s}{1:s}".format(logStr,'logging.exception!')) 
        logging.exception('')  
    else:
        logger.debug("{0:s}{1:s}".format(logStr,'No Exception.')) 
        sys.exit(0)
    finally:
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 


        sys.exit(0)

